/*  This file is part of hpafs (Hidden Protected Area FileSystem). hpafs is a
 *  FUSE (Filesystem in USErspace).
 *
 *  Copyright (C) 2006, 2007 Paul Bolle <pebolle@tiscali.nl>
 *
 *  hpafs is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  hpafs is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <linux/hdreg.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "beer.h"
#include "dos.h"

#define LBA                     0x40
#define SLAVE                   0       /* see setmax.c for background */

#ifndef WIN_READ_NATIVE_MAX
#define WIN_READ_NATIVE_MAX     0xf8
#endif

static char * progname;
static char * device;
static unsigned int native_max;

int hpafs_device_identify(int fd, unsigned char *args);

int
hpafs_device_native_max(unsigned int * max)
{
    int fd, ret;
    unsigned char args[7];
    int i;

    fd = open(device, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "%s: couldn't open%s!\n", progname, device);
        return 0;
    }

    for (i=0; i<7; i++)
        args[i] = 0;
    args[0] = WIN_READ_NATIVE_MAX;
    args[6] = (SLAVE ? 0x10 : 0) | LBA;
    if (ioctl(fd, HDIO_DRIVE_CMD_AEB, &args)) {
        perror("HDIO_DRIVE_CMD_AEB failed WIN_READ_NATIVE_MAX");
        return 0;
    }

    *max = ((args[6] & 0xf) << 24) + (args[5] << 16) + (args[4] << 8) + args[3];

    /* return value is maxlba -1, so add 1 */
    *max += 1;

    ret = close(fd);
    if (ret == -1) {
        fprintf(stderr, "%s: couldn't close %s!\n", progname, device);
        return 0;
    }

    return 1;
}

int
hpafs_device_capacity(unsigned int * cap){
    int fd, ret;
    unsigned char args[4+512];
    struct hd_driveid *id = (struct hd_driveid *)&args[4];

    fd = open(device, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "%s: couldn't open%s!\n", progname, device);
        return 0;
    }

    ret = hpafs_device_identify(fd, &args[0]);
    if (!ret) {
        fprintf(stderr, "%s: couldn't identify %s!\n", progname, device);
        return 0;
    }

    *cap = id->lba_capacity;

    ret = close(fd);
    if (ret == -1) {
        fprintf(stderr, "%s: couldn't close %s!\n", progname, device);
        return 0;
    }

    return 1;
}

static void
hpafs_usage(const char *progname, struct fuse_operations * oper)
{
    const char *hpafs_help[] = { progname, "-ho", NULL };

    fprintf(stderr,
"usage: %s device mountpoint [FUSE options]\n"
/*"\n"
"HPAFS Options:\n"
"                           none\n"*/
"\n", progname);
    fuse_main(2, (char **) hpafs_help, oper);
    exit(0);
}

void
hpafs_test_device(void){
    struct stat *buf;
    int ret;

    buf = (struct stat *) malloc (sizeof(struct stat));
    if (buf == NULL) {
        fprintf(stderr, "%s out of memory!\n", progname);
        exit (1);
    }

    if ((ret = stat(device, buf)) < 0) { 
        fprintf(stderr, "%s: stat(%s) failed!\n", progname, device);
        exit (1);
    }

    if (!(ret = S_ISBLK (buf->st_mode))) {
        fprintf(stderr, "%s: %s is not a block device!\n", progname, device);
        exit(1);
    };
 
    if (buf) free(buf);
}

int
hpafs_device_identify(int fd, unsigned char *args){

   *args = WIN_IDENTIFY;
   *(args+1) = 0;
   *(args+2) = 0;
   *(args+3) = 1;

   if (ioctl(fd, HDIO_DRIVE_CMD, args)) {
        *args = WIN_PIDENTIFY;
        if (ioctl(fd, HDIO_DRIVE_CMD, args)) {
            fprintf(stderr, "%s: ioctl failed!\n", progname);
            return 0;
        }
    }

    return 1;
}

void
hpafs_hpa_supported(void){
    int fd, ret;
    unsigned char args[4+512];
    struct hd_driveid *id = (struct hd_driveid *)&args[4];

    fd = open(device, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "%s: couldn't open %s!\n", progname, device);
        exit(1);
    }

    hpafs_device_identify(fd, &args[0]);

    if ((!id->command_set_1 & 0x400) && (!id->cfs_enable_1 & 0x400)){
        fprintf(stderr, "%s: Host Protected Area not supported on %s!\n", progname, device);
        exit (1);
    }

    ret = close(fd);
    if (ret == -1) {
        fprintf(stderr, "%s: couldn't close %s!\n", progname, device);
        exit (1);
    }

    return;
}

static int
hpafs_getattr(const char *path, struct stat *stbuf)
{
    unsigned int cap;
    int fd;
    int res = 0;
    int is_path = 0;
    beer_t beer;
    direntry_t *dos;
    off_t off;
    int offset;
    int i, ret;
    char *dos_path;

    if(!hpafs_device_capacity(&cap)) {
        fprintf(stderr, "%s: couldn't get capacity of %s\n", progname, device);
        return -ENOENT;
    }

    if (cap != native_max) {
        fprintf(stderr, "%s: capacity = %d; native max address = %d\n", device,
                cap, native_max);
        fprintf(stderr, "capacity doesn't equal native max address\n");
        return -ENOENT;
    }

    fd = open(device, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "%s: couldn't open%s!\n", progname, device);
        return -ENOENT;
    }

    if ((off = lseek(fd, -512, SEEK_END) == (off_t)-1)) {
        perror ("hpafs_getattr lseek");
        return -ENOENT;
    };

    read_beer(&beer, fd, 1);

    dos = (direntry_t *) malloc (beer.num_entries * sizeof (direntry_t));
    if (dos == NULL) {
        fprintf (stderr, "malloc failed\n");
        return -ENOENT;
    }

    for (i = 0; i < beer.num_entries; i++) {
        offset = 512 * (1 + (i + 2) / 8) - 64 * ((i + 2 ) % 8);

        if ((off = lseek(fd, -offset, SEEK_END)) == (off_t)-1) {
            perror ("hpafs_getattr lseek");
            if (dos) free(dos);
            return -ENOENT;
        }
    
        read_dos(dos+i, fd, 1);
    }

    memset(stbuf, 0, sizeof(struct stat));
    if(strcmp(path, "/") == 0) {
        stbuf->st_mode = S_IFDIR | 0755;
        stbuf->st_nlink = 2;
        is_path = 1;
    }

    for (i = 0; (i < beer.num_entries) && (is_path == 0); i++) {
        dos_path = (char *) malloc (sizeof (char) * (1 + strlen((dos+i)->id_string) + 1));
        if (dos_path == NULL) {
             fprintf (stderr, "malloc failed\n");
             break;
        }
        strcpy (dos_path, "/"); 
        strcpy (dos_path+1, (dos+i)->id_string); 
        if(strcmp(path, dos_path) == 0) {
            stbuf->st_mode = S_IFREG | 0444;
            stbuf->st_nlink = 1;
            stbuf->st_size = (dos+i)->service_area_size * 512;
            stbuf->st_blocks = (dos+i)->service_area_size;
            is_path = 1;
        }
        free (dos_path);
    }

    if (!is_path)
        res = -ENOENT;

    free(dos);

    ret = close(fd);
    if (ret == -1) {
        fprintf(stderr, "%s: couldn't close %s!\n", progname, device);
        res = -ENOENT;
    }

    return res;
}

static int
hpafs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
              off_t offset, struct fuse_file_info *fi)
{
    (void) offset;
    (void) fi;
    unsigned int cap;
    beer_t beer;
    direntry_t *dos;
    off_t off;
    int off_n;
    int fd;
    int ret, i;
    char *dos_path;

    if(!hpafs_device_capacity(&cap)) {
        fprintf(stderr, "%s: couldn't get capacity of %s\n", progname, device);
        return -ENOENT;
    }

    if (cap != native_max) {
        fprintf(stderr, "%s: capacity = %d; native max address = %d\n", device,
                cap, native_max);
        fprintf(stderr, "capacity doesn't equal native max address\n");
        return -ENOENT;
    }

    if(strcmp(path, "/") != 0)
        return -ENOENT;

    fd = open(device, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "%s: couldn't open%s!\n", progname, device);
        return -ENOENT;
    }

    if ((off = lseek(fd, -512, SEEK_END) == (off_t)-1)) {
        perror ("hpafs_readdir lseek");
        return -ENOENT;
    };

    read_beer(&beer, fd, 1);

    dos = (direntry_t *) malloc (beer.num_entries * sizeof (direntry_t));
    if (dos == NULL) {
        fprintf (stderr, "malloc failed\n");
        return -ENOENT;
    }

    for (i = 0; i < beer.num_entries; i++) {
        off_n = 512 * (1 + (i + 2) / 8) - 64 * ((i + 2 ) % 8);

        if ((off = lseek(fd, -off_n, SEEK_END)) == (off_t)-1) {
            perror ("hpafs_readdir lseek");
            return -ENOENT;
        }
    
        read_dos(dos+i, fd, 1);
    }

    filler(buf, ".", NULL, 0);
    filler(buf, "..", NULL, 0);

    for (i = 0; i < beer.num_entries; i++) {
        dos_path = (char *) malloc (sizeof (char) * (1 + strlen((dos+i)->id_string) + 1));
        if (dos_path == NULL) {
            fprintf (stderr, "malloc failed\n");
            return -ENOENT;
        }
        strcpy (dos_path, "/"); 
        strcpy (dos_path+1, (dos+i)->id_string); 
        filler(buf, dos_path + 1, NULL, 0);
        free(dos_path);
    }

    if (dos) free(dos);

    ret = close(fd);
    if (ret == -1) {
        fprintf(stderr, "%s: couldn't close %s!\n", progname, device);
        return -ENOENT;
    }

    return 0;
}

static int hpafs_open(const char *path, struct fuse_file_info *fi)
{
    unsigned int cap;
    beer_t beer;
    direntry_t *dos;
    off_t off;
    int off_n;
    int fd;
    int ret, i;
    char *dos_path;

    int known_path = 0;

    if(!hpafs_device_capacity(&cap)) {
        fprintf(stderr, "%s: couldn't get capacity of %s\n", progname, device);
        return -ENOENT;
    }

    if (cap != native_max) {
        fprintf(stderr, "%s: capacity = %d; native max address = %d\n", device,
                cap, native_max);
        fprintf(stderr, "capacity doesn't equal native max address\n");
        return -ENOENT;
    }

    fd = open(device, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "%s: couldn't open%s!\n", progname, device);
        return -ENOENT;
    }

    if ((off = lseek(fd, -512, SEEK_END) == (off_t)-1)) {
        perror ("hpafs_getattr lseek");
        return -ENOENT;
    };

    read_beer(&beer, fd, 1);

    dos = (direntry_t *) malloc (beer.num_entries * sizeof (direntry_t));
    if (dos == NULL) {
        fprintf (stderr, "malloc failed\n");
        return -ENOENT;
    }

    for (i = 0; i < beer.num_entries; i++) {
        off_n = 512 * (1 + (i + 2) / 8) - 64 * ((i + 2 ) % 8);

        if ((off = lseek(fd, -off_n, SEEK_END)) == (off_t)-1) {
            perror ("hpafs_getattr lseek");
            if (dos) free(dos);
            return -ENOENT;
        }
    
        read_dos(dos+i, fd, 1);
    }

    for (i = 0; (i < beer.num_entries) && (known_path == 0); i++) {
        dos_path = (char *) malloc (sizeof (char) * (1 + strlen((dos+i)->id_string) + 1));
        if (dos_path == NULL) {
             fprintf (stderr, "malloc failed\n");
             break;
        }
        strcpy (dos_path, "/"); 
        strcpy (dos_path+1, (dos+i)->id_string); 
        if(strcmp(path, dos_path) == 0)
            known_path = 1;
        free (dos_path);
    }

    if (dos) free(dos);

    ret = close(fd);
    if (ret == -1) {
        fprintf(stderr, "%s: couldn't close %s!\n", progname, device);
        return -ENOENT;
    }

    if (!known_path)
        return -ENOENT;

    if((fi->flags & 3) != O_RDONLY)
        return -EACCES;

    return 0;
}

static int hpafs_read(const char *path, char *buf, size_t size, off_t offset,
                      struct fuse_file_info *fi)
{
    size_t len;
    (void) fi;
    unsigned int cap;
    beer_t beer;
    direntry_t *dos;
    off_t off;
    int off_n;
    int fd;
    int ret, i, j;
    char *dos_path;
    ssize_t res;

    int known_path = 0;

    if(!hpafs_device_capacity(&cap)) {
        fprintf(stderr, "%s: couldn't get capacity of %s\n", progname, device);
        return -ENOENT;
    }

    if (cap != native_max) {
        fprintf(stderr, "%s: capacity = %d; native max address = %d\n", device,
                cap, native_max);
        fprintf(stderr, "capacity doesn't equal native max address\n");
        return -ENOENT;
    }

    fd = open(device, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "%s: couldn't open%s!\n", progname, device);
        return -ENOENT;
    }

    if ((off = lseek(fd, -512, SEEK_END) == (off_t)-1)) {
        perror ("hpafs_getattr lseek");
        return -ENOENT;
    };

    read_beer(&beer, fd, 1);

    dos = (direntry_t *) malloc (beer.num_entries * sizeof (direntry_t));
    if (dos == NULL) {
        fprintf (stderr, "malloc failed\n");
        return -ENOENT;
    }

    for (i = 0; i < beer.num_entries; i++) {
        off_n = 512 * (1 + (i + 2) / 8) - 64 * ((i + 2 ) % 8);

        if ((off = lseek(fd, -off_n, SEEK_END)) == (off_t)-1) {
            perror ("hpafs_getattr lseek");
            if (dos) free(dos);
            return -ENOENT;
        }
    
        read_dos(dos+i, fd, 1);
    }

    for (i = 0; (i < beer.num_entries) && (known_path == 0); i++) {
        dos_path = (char *) malloc (sizeof (char) * (1 + strlen((dos+i)->id_string) + 1));
        if (dos_path == NULL) {
             fprintf (stderr, "malloc failed\n");
             break;
        }
        strcpy (dos_path, "/"); 
        strcpy (dos_path+1, (dos+i)->id_string); 
        if(strcmp(path, dos_path) == 0) {
            known_path = 1;
            j = i;
        }
        free (dos_path);
    }

    if (!known_path)
        return -ENOENT;

    len =  (dos+j)->service_area_size * 512;
    if (offset < len) {
        if (offset + size > len)
            size = len - offset;
        if ((off = lseek(fd,
            (off_t) ((dos+j)->service_area_start * 512) + offset,
            SEEK_SET) == (off_t)-1)) {
            fprintf(stderr, "%s: hpafs_read lseek\n", (dos+j)->id_string);
            return -ENOENT; /* -EIO? */
        };
        res = read(fd, buf, size);
        if ((size_t) res != size)
            fprintf(stderr, "%s: short read\n", (dos+j)->id_string);
    } else
        size = 0;

    if (dos) free(dos);

    ret = close(fd);
    if (ret == -1) {
        fprintf(stderr, "%s: couldn't close %s!\n", progname, device);
        return -ENOENT;
    }

    return size;
}

static struct fuse_operations hpafs_oper = {
    .getattr    = hpafs_getattr,
    .readdir    = hpafs_readdir,
    .open       = hpafs_open,
    .read       = hpafs_read,
};

int main(int argc, char *argv[])
{
    int i;
    unsigned int cap;

    /* http://cvs.sourceforge.net/viewcvs.py/fuse/sshfs/sshfs.c?rev=1.28&view=markup */
    for (i=1; i < argc; i++) {
        if (strcmp(argv[i], "-h") == 0)
            hpafs_usage(argv[0], &hpafs_oper);
    }

    if (argc < 3) {
        fprintf(stderr, "%s: missing device (\"mountpath\") or mountpoint\n",
                argv[0]);
        fprintf(stderr, "see '%s -h' for usage\n", argv[0]);
        exit(1);
    }

    progname = argv[0];
    device = argv[1];

    hpafs_test_device();

    hpafs_hpa_supported();

    if (!hpafs_device_native_max(&native_max)) {
        fprintf(stderr, "%s: couldn't determine native max address of %s\n",
                progname, device);
        exit(1);
    }

    if(!hpafs_device_capacity(&cap)) {
        fprintf(stderr, "%s: couldn't get capacity of %s\n", progname, device);
        exit(1);
    }

    if (cap != native_max) {
        fprintf(stderr, 
                "%s: capacity of %s is %u, native max address is %u.\n"
                "cannot create %s with Host Protected Area enabled.\n",
                progname, device, cap, native_max, progname);
        exit(1);
    }

    /* drop argv[1] (device a.k.a. "mountpath"); device points to old argv[1] */
    for (i=1; i < argc; i++) 
        argv[i] = argv[i+1];
    argc--;

    return fuse_main(argc, argv, &hpafs_oper);
}
